/*
 * Copyright(c) 1997, Space Science and Engineering Center, UW-Madison
 * Refer to "McIDAS Software Acquisition and Distribution Policies"
 * in the file  mcidas/data/license.txt
 */

/**** $Id: mct_winx.c,v 1.13 2003/12/17 20:38:58 davep Exp $ ****/

#if defined(__GNUC__) && defined(__STRICT_ANSI__)
/* turn off gcc '-pedantic' flag so we get 'struct sigaction' defined */
#undef __STRICT_ANSI__
#if defined(__sun)
#define __EXTENSIONS__
#endif
#endif

#include <sys/types.h>
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <sys/signal.h>
#include <assert.h>

#include <X11/Xlib.h>
#include <X11/Xproto.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>	/* for XtNinput */

#include "mcidasp.h"
#include "m0glue.h"
#include "m0key.h"
#include "mcterm.h"
#include "mct_win.h"
#include "mct_ebuf.h"
#include "mct_obuf.h"


static void TWfini(void);
static void TWbeep(int freq, int ms);

/*
 * Fallback resources
 */


/*
 * ! the below font is also known as 10x20
 *
 * *font: -misc-fixed-medium-r-normal--20-*-*-*-c-*-iso8859-1
 *
 * ! the below font is also known as 9x15bold
 *
 * *font: -misc-fixed-bold-r-normal--15-*-*-*-c-*-iso8859-1
 *
 * Note that we MUST use a character cell font (indicated by the -c-
 * in the above descriptions) for the mcterm widget to work properly.
 */

static String fallback_resources[] = {
    "*fontListHead: "
		    "5x8,"
		    "6x9,"
		    "6x10,"
		    "6x12,"
		    "6x13bold,"
		    "7x13bold,"
		    "8x13bold,"
		    "7x14bold,"
		    ,

    "*font: 9x15bold",

    "*fontListTail: "
		    "8x16,"
		    "10x20,"
		    "12x24,"
		    ,

    "*background: black",
    "*foreground: white",
    "*foregroundBright: white",

    "Mctext.geometry: 85x26+300+400",

    NULL,
};

/*
 * allow -tgeometry to be an alias for -geometry
 */

static XrmOptionDescRec options[] = {
{"-tgeometry",   ".geometry",    XrmoptionSepArg,    (XtPointer) NULL},

{"-fontListHead","*fontListHead",XrmoptionSepArg,    (XtPointer) NULL},
{"-fontListTail","*fontListTail",XrmoptionSepArg,    (XtPointer) NULL},

{"-foregroundBright",  "*foregroundBright",  XrmoptionSepArg, (XtPointer) NULL},
{"-fgBright",          "*foregroundBright",  XrmoptionSepArg, (XtPointer) NULL},

{"-foregroundBrightBlack",  "*foregroundBrightBlack",  XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightRed",    "*foregroundBrightRed",    XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightGreen",  "*foregroundBrightGreen",  XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightYellow", "*foregroundBrightYellow", XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightBlue",   "*foregroundBrightBlue",   XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightMagenta","*foregroundBrightMagenta",XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightCyan",   "*foregroundBrightCyan",   XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundBrightWhite",  "*foregroundBrightWhite",  XrmoptionSepArg, (XtPointer) NULL},

{"-fgBBlack",          "*foregroundBrightBlack",  XrmoptionSepArg, (XtPointer) NULL},
{"-fgBRed",            "*foregroundBrightRed",    XrmoptionSepArg, (XtPointer) NULL},
{"-fgBGreen",          "*foregroundBrightGreen",  XrmoptionSepArg, (XtPointer) NULL},
{"-fgBYellow",         "*foregroundBrightYellow", XrmoptionSepArg, (XtPointer) NULL},
{"-fgBBlue",           "*foregroundBrightBlue",   XrmoptionSepArg, (XtPointer) NULL},
{"-fgBMagenta",        "*foregroundBrightMagenta",XrmoptionSepArg, (XtPointer) NULL},
{"-fgBCyan",           "*foregroundBrightCyan",   XrmoptionSepArg, (XtPointer) NULL},
{"-fgBWhite",          "*foregroundBrightWhite",  XrmoptionSepArg, (XtPointer) NULL},

{"-foregroundNormalBlack",  "*foregroundNormalBlack",  XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalRed",    "*foregroundNormalRed",    XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalGreen",  "*foregroundNormalGreen",  XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalYellow", "*foregroundNormalYellow", XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalBlue",   "*foregroundNormalBlue",   XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalMagenta","*foregroundNormalMagenta",XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalCyan",   "*foregroundNormalCyan",   XrmoptionSepArg, (XtPointer) NULL},
{"-foregroundNormalWhite",  "*foregroundNormalWhite",  XrmoptionSepArg, (XtPointer) NULL},

{"-fgNBlack",          "*foregroundNormalBlack",  XrmoptionSepArg, (XtPointer) NULL},
{"-fgNRed",            "*foregroundNormalRed",    XrmoptionSepArg, (XtPointer) NULL},
{"-fgNGreen",          "*foregroundNormalGreen",  XrmoptionSepArg, (XtPointer) NULL},
{"-fgNYellow",         "*foregroundNormalYellow", XrmoptionSepArg, (XtPointer) NULL},
{"-fgNBlue",           "*foregroundNormalBlue",   XrmoptionSepArg, (XtPointer) NULL},
{"-fgNMagenta",        "*foregroundNormalMagenta",XrmoptionSepArg, (XtPointer) NULL},
{"-fgNCyan",           "*foregroundNormalCyan",   XrmoptionSepArg, (XtPointer) NULL},
{"-fgNWhite",          "*foregroundNormalWhite",  XrmoptionSepArg, (XtPointer) NULL},

{"-backgroundBlack",  "*backgroundBlack",  XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundRed",    "*backgroundRed",    XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundGreen",  "*backgroundGreen",  XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundYellow", "*backgroundYellow", XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundBlue",   "*backgroundBlue",   XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundMagenta","*backgroundMagenta",XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundCyan",   "*backgroundCyan",   XrmoptionSepArg, (XtPointer) NULL},
{"-backgroundWhite",  "*backgroundWhite",  XrmoptionSepArg, (XtPointer) NULL},

{"-bgBlack",          "*backgroundBlack",  XrmoptionSepArg, (XtPointer) NULL},
{"-bgRed",            "*backgroundRed",    XrmoptionSepArg, (XtPointer) NULL},
{"-bgGreen",          "*backgroundGreen",  XrmoptionSepArg, (XtPointer) NULL},
{"-bgYellow",         "*backgroundYellow", XrmoptionSepArg, (XtPointer) NULL},
{"-bgBlue",           "*backgroundBlue",   XrmoptionSepArg, (XtPointer) NULL},
{"-bgMagenta",        "*backgroundMagenta",XrmoptionSepArg, (XtPointer) NULL},
{"-bgCyan",           "*backgroundCyan",   XrmoptionSepArg, (XtPointer) NULL},
{"-bgWhite",          "*backgroundWhite",  XrmoptionSepArg, (XtPointer) NULL},

/*
 * text color options in terms of TCOL= values
 */

{"-tcolbg0",   "*backgroundBlack",         XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg16",  "*backgroundBlue",          XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg32",  "*backgroundGreen",         XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg48",  "*backgroundCyan",          XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg64",  "*backgroundRed",           XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg80",  "*backgroundMagenta",       XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg96",  "*backgroundYellow",        XrmoptionSepArg, (XtPointer) NULL},
{"-tcolbg112", "*backgroundWhite",         XrmoptionSepArg, (XtPointer) NULL},

{"-tcolfg0",   "*foregroundNormalBlack",   XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg1",   "*foregroundNormalBlue",    XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg2",   "*foregroundNormalGreen",   XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg3",   "*foregroundNormalCyan",    XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg4",   "*foregroundNormalRed",     XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg5",   "*foregroundNormalMagenta", XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg6",   "*foregroundNormalYellow",  XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg7",   "*foregroundNormalWhite",   XrmoptionSepArg, (XtPointer) NULL},

{"-tcolfg8",   "*foregroundBrightBlack",   XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg9",   "*foregroundBrightBlue",    XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg10",  "*foregroundBrightGreen",   XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg11",  "*foregroundBrightCyan",    XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg12",  "*foregroundBrightRed",     XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg13",  "*foregroundBrightMagenta", XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg14",  "*foregroundBrightYellow",  XrmoptionSepArg, (XtPointer) NULL},
{"-tcolfg15",  "*foregroundBrightWhite",   XrmoptionSepArg, (XtPointer) NULL},
};



static Atom wm_protocols;
static Atom wm_delete_window;
static Widget toplevel;
static Widget textout;
static XtAppContext app_context;

static int TWrows;	/* current number of rows in textout */
static int TWcols;	/* current number of columns in textout */

static M0buf *TWbuf;	/* output to the widget is buffered here */

/* safe signal handler installer */

static void (*mysignal(int signo, void (*func)(int signo)))(int signo);

/* signal handlers - these simply turn on their bits in hitbits */

enum sighits { HIT_WINCH = 1, HIT_TSTP = 2, HIT_INTERRUPT = 4 };

static volatile sig_atomic_t hitbits;

static void handle_signal(int signo);

/* signal handler - this simply turns on bits in hitbits */

static void
handle_signal(int signo)
{
	switch(signo)
	{
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		hitbits |= HIT_INTERRUPT;
		break;
	}
}


/* install signal handlers safely */

static void (*mysignal(int signo, void (*func)(int signo)))(int signo)
{
	struct sigaction act, oact;

	act.sa_handler = func;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
		act.sa_flags |= SA_INTERRUPT;
#endif
	} else {
#ifdef SA_RESTART
		act.sa_flags |= SA_RESTART;
#endif
	}

	if (sigaction(signo, &act, &oact) == -1)
	{
#ifdef DEBUG
		char buf[128];
		sprintf(buf, "signal %d, errno %d, desc %s\n", signo,
		        errno, strerror(errno));
		write(STDERR_FILENO, buf, strlen(buf));
#endif
		return SIG_ERR;
	}
	return oact.sa_handler;
}

/*--------------------------------------------------------------------
 |  The key_cb callback is invoked whenever the user presses a key in
 |  the McTerm widget. The call_data argument contains a pointer to
 |  the corresponding XKeyEvent.
 |
 |  The client_data is a pointer to the file descriptor of the pipe
 |  to the child process.
 +--------------------------------------------------------------------*/

static char key_token[] = "blah";

static void
key_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *token = (char *) client_data;
	XKeyEvent *event = (XKeyEvent *) call_data;

	M0Key k;
	int rc;

	assert(w);
	assert(token == key_token);

	rc = M0XKeyToM0Key(event, &k);

	if (rc < 0)
	{
		int len;
		char buf[10];
		KeySym sym;
		char *name;

		static char prefix[] = "unknown key";

		static M0buf *desc;

		if(!desc)
			desc = M0bufalloc();

		len = XLookupString(event, buf, sizeof buf, &sym, NULL);

		M0bufcpy(desc, prefix, sizeof prefix - 1);

		name = XKeysymToString(sym);
		if (name)
		{
			M0bufcat(desc, " (", 2);
			M0bufcat(desc, name, strlen(name));
			M0bufcat(desc, ")", 1);
		}

		if (len)
		{
			int i;
			M0bufcat(desc, " ``", 3);
			for (i = 0; i < len; i++)
			{
				if (isprint(buf[i]))
					M0bufcat(desc, buf+i, 1);
				else
					M0bufcat(desc, " ", 1);
			}
			M0bufcat(desc, "''", 2);
		}

		{
			char buf[64];
			sprintf(buf, " keysym 0x%lX ", (long) sym);
			M0bufcat(desc, buf, strlen(buf));
		}

		OBwrite(TCOLOR_FOREGROUND_BRIGHT_YELLOW, M0bufptr(desc));
		TWbeep(-1,-1);
	}
	if (rc <= 0)
	{
		return;
	}

	EBgeteditor()->EBhandle(k, 0);
}

/*--------------------------------------------------------------------
 |  The mouse_cb callback is invoked whenever the user presses the CTRL-Left mouse button in
 |  the McImage widget. The call_data argument contains a pointer to
 |  the corresponding XButtonEvent.
 |
 |  The client_data is a pointer to the file descriptor of the pipe
 |  to the child process.
 +--------------------------------------------------------------------*/
static char mouse_token[] = "blah";

static void
mouse_cb(Widget w, XtPointer client_data, XtPointer call_data)
  {
  /* text is passed to the edit buffer by sending it through the active editor */
    M0Key k;            /* data passing structure for each key press */
    char tbuffer[128];  /* buffer used to hold text data to send     */
    int i,length;       /* temp counter and string length            */
    
    int PasteFlag;      /* set to 1(true) if pasting data into the EBuffer */

  char *token = (char *) client_data;
  XButtonEvent *event = (XButtonEvent *) call_data;

  /* this event is always and only triggered by a Lat/Lon request */
  /* verify that the values in UC are valid nav values -- if so dump to the Edit Line */
  
  int Lat=Mcluc(UC_CURSOR_LATITUDE);
  int Lon=Mcluc(UC_CURSOR_LONGITUDE);
  
  int LatMin,LatSec,LonMin,LonSec;

  float Lat_real;  /* Used to compute lat in decimal form */
  float Lon_real;  /* Used to compute lon in decimal form */

  
  if(Lat==MCMISSING4 || Lon==MCMISSING4 || !Mcluc(UC_UPDATE_CURSOR_INFO))
    {
    /* the Mc-- calls are not hooked up -- so need to use obuffer ?? */
    /*if(Mcluc(UC_UPDATE_CURSOR_INFO)) 
      Mceprintf("Unable to Navigate that point.\n");
    else
      Mceprintf("Cursor Tracking has not been turned on.\n"); */
    TWbeep(-1,-1);
    return;
    }

  /* step one -- build the string to send to the editor */
  /* access UC to get the LAT and LON values */
    LatSec=Lat%100;
    LatMin=(Lat/100)%100;
    Lat=Lat/10000;
    LonSec=Lon%100;
    LonMin=(Lon/100)%100;
    Lon=Lon/10000;
    
    if(LatMin<0) 
      {
      LatMin=-LatMin;
      LatSec=-LatSec;
      }
    if(LonMin<0) 
      {
      LonMin=-LonMin;
      LonSec=-LonSec;
      }

/*    
   Read UC word 17 to determine whether DMS or decimal format should
   be used, if UC word 17 does not equal zero, then convert DMS to 
   decimal. 
*/
    if(Mcluc(17) != 0) 
    {
	Lat_real = LatSec / 60.0;
	Lat_real = (LatMin + Lat_real) / 60.0;
	Lat_real = Lat + Lat_real;

	Lon_real = LonSec / 60.0;
	Lon_real = (LonMin + Lon_real) / 60.0;
	Lon_real = Lon + Lon_real;

	sprintf(tbuffer," %f %f",Lat_real,Lon_real);
    }
    else
    {
	sprintf(tbuffer," %i:%02i:%02i %i:%02i:%02i",Lat,LatMin,LatSec,Lon,LonMin,LonSec);
    }
  
  /* set up a loop to "paste" each character in the buffer to the editor */
    length=strlen(tbuffer);       /* how many characters? */ 
    PasteFlag=1;                  /* yup, we're pasting   */
    for(i=0;i<length;i++)
      {
      k.name= (int)tbuffer[i];
      k.mod= M0KEYMOD_NONE;
      EBmcidas()->EBhandle(k, PasteFlag);   
      }
  }


/*--------------------------------------------------------------------
 |  The paste_cb callback is invoked whenever the user pastes a byte in
 |  the McTerm widget. The call_data argument contains a pointer to
 |  the corresponding character.
 |
 |  The client_data is a pointer to the file descriptor of the pipe
 |  to the child process.
 +--------------------------------------------------------------------*/

static void
EWchar(char c, int pasteflag)
{
	M0Key k;

#define myctrl(x)	((x)&037)
	switch(c)
	{
	case myctrl('H'):
		k.name = M0KEYNAME_BACKSPACE;
		k.mod = M0KEYMOD_NONE;
		EBgeteditor()->EBhandle(k, pasteflag);
		break;
	case myctrl('I'):
		/* turn tabs to spaces */
		k.name = SPACE;
		k.mod = M0KEYMOD_NONE;
		EBgeteditor()->EBhandle(k, pasteflag);
		break;
	case myctrl('M'):
		k.name = M0KEYNAME_ENTER;
		k.mod = M0KEYMOD_NONE;
		EBgeteditor()->EBhandle(k, pasteflag);
		break;
	default:
		k.name = (int) c;
		k.mod = M0KEYMOD_NONE;

		if(0 <= k.name && k.name < 0x20)
		{
			/* turn into an upper-case letter */

			k.name += 0x40;
			k.mod = M0KEYMOD_CONTROL;
		}

		EBgeteditor()->EBhandle(k, pasteflag);
		break;
	}
}

static char paste_token[] = "blah";

static void
paste_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *token = (char *) client_data;
	char *data = (char *) call_data;

	/* If MCPASTE=NONE, do nothing */
	char * mcpaste_env = getenv("MCPASTE");
	if (mcpaste_env != NULL && strcmp(mcpaste_env, "NONE") == 0) {
	    return;
	}

	assert(w);
	assert(token == paste_token);

	EWchar(*data, 1);
}

/*--------------------------------------------------------------------
 |  The main body initializes the X Toolkit, creates the sub-process
 |  and installs the callback and the input handler.
 +--------------------------------------------------------------------*/

static void
exit_handler(Widget w, XtPointer client_data, XEvent *event, Boolean *continue_to_dispatch)
{
	Widget token = (Widget)client_data;

	assert(token == w);
	assert(continue_to_dispatch);

	if (event->type == ClientMessage
	 && event->xclient.message_type == wm_protocols
	 && event->xclient.format == 32
	 && event->xclient.data.l[0] == wm_delete_window)
	{
		TWfini();
		exit(0);
	}
}


/*
 * Change the font in the given textout widget.
 */

static void
set_font(Widget textout, const char *name)
{
	XtVaSetValues(textout,
		XtVaTypedArg, XtNfont, XtRString, name, strlen(name)+1,
		NULL);
}

/*
 * Appends the given font name to the given vector.
 * Returns the new vector.
 * Eliminates leading and trailing white space from the name,
 * and skips empty names.
 */

static char **
append_font_name(char **vec, const char *name)
{
	char  *work;
	char **ans;

	if (!name || !*name)
		return vec;

	/* skip leading white space */
	while(*name && isspace(*name))
		name++;

	/* skip trailing white space */
	work = fsalloc(name, (FsLen) strlen(name));
	assert(work);

	/* if it's not empty, add it to the list */

	ans = (*work) ? VecAdd(vec, work) : vec;

	free(work);

	return ans;
}

/*
 * Parses the given comma-separated list of fonts and appends them
 * to the given vector.  Returns the new vector.
 */

static char **
append_font_list(char **vec, const char *list)
{
	char *work;
	char *beg;
	char *end;

	if (!list || !*list)
		return vec;

	beg = work = stralloc(list, (char *)0);

	while((end = strchr(beg, ',')) != NULL)
	{
		*end = '\0';
		vec = append_font_name(vec, beg);
		beg = end+1;
	}
	vec = append_font_name(vec, beg);

	free(work);
	return vec;
}

/*
 * Return the list of fonts specified in the fontListHead,
 * font, and fontListTail resources.
 *
 * Store the index of the initial font (the one specified in the font
 * resource) in cur_font.
 *
 * This routine assumes that the font resource is set.  In order to
 * guarantee this is the case, the application must have a fallback
 * setting for this resource.
 */

static char **
get_font_list(Widget toplevel, int *cur_font)
{
	char **vec;

	typedef struct
	{
	String	fonthead;
	String	fontname;
	String	fonttail;
	}	AppData, *AppDataPtr;

	static XtResource resources[] = {
	{	"font", "font", XtRString,
	        sizeof(String), XtOffset(AppDataPtr, fontname),
	        XtRString, NULL },
	{	"fontListHead", "FontListHead", XtRString,
	        sizeof(String), XtOffset(AppDataPtr, fonthead),
	        XtRString, NULL },
	{	"fontListTail", "FontListTail", XtRString,
	        sizeof(String), XtOffset(AppDataPtr, fonttail),
	        XtRString, NULL },
	};

	static AppData data;

	XtVaGetApplicationResources(toplevel, &data,
			    resources, XtNumber(resources), NULL);

	vec = VecNew();
	assert(vec);

	if (data.fonthead)
	{
		vec = append_font_list(vec, data.fonthead);
		assert(vec);
	}

	assert(data.fontname);
	assert(cur_font);
	*cur_font = VecLen(vec);
	vec = VecAdd(vec, data.fontname);

	if (data.fonttail)
	{
		vec = append_font_list(vec, data.fonttail);
		assert(vec);
	}

	return vec;
}

static char **font_list;
static int cur_font;


static void TWprev_font(void)
{
	assert(font_list);

	if (cur_font == 0)
	{
		WINerror("at head of font list");
		return;
	}
	cur_font--;
	set_font(textout, font_list[cur_font]);
}

static void TWnext_font(void)
{
	assert(font_list);

	if (cur_font + 1 >= VecLen(font_list))
	{
		WINerror("at tail of font list");
		return;
	}
	cur_font++;
	set_font(textout, font_list[cur_font]);
}


static char *TWinit(int *argc, char **argv)
{
	assert(argc);
	assert(argv);

	/*
	 * We buffer output to the widget in order to minimize the
	 * amount of X activity it has to do.  It updates its window
	 * after each XMcTermWrite(), so the fewer the calls to
	 * XMcTermWrite(), the less X activity the widget will have to
	 * do.
	 */

	TWbuf = M0bufalloc();

	mysignal(SIGINT, handle_signal);
	mysignal(SIGTERM, handle_signal);
	mysignal(SIGHUP, handle_signal);

	toplevel = XtVaAppInitialize(&app_context, "Mctext",
				    options, XtNumber(options),
				    argc, argv,
				    fallback_resources,
				    XtNinput, (XtArgVal) True,
				    XtNallowShellResize, (XtArgVal) True,
				    NULL);

	font_list = get_font_list(toplevel, &cur_font);

#if 0
{
	int i;

	for(i = 0; i < VecLen(font_list); i++)
	{
		printf("font_list[%d] = %s\n", i, font_list[i]);
	}
	printf("cur_font = %d\n", cur_font);
	fflush(stdout);
}
#endif

	wm_protocols = XInternAtom(XtDisplay(toplevel),
				   "WM_PROTOCOLS", False);
	wm_delete_window = XInternAtom(XtDisplay(toplevel),
				       "WM_DELETE_WINDOW", False);

	XtAddEventHandler(toplevel, NoEventMask, True, exit_handler,
			  (XtPointer) toplevel);

	textout = XtVaCreateManagedWidget("mcterm", xmcTermWidgetClass,
					   toplevel, NULL);

	XtAddCallback(textout, XtNkeyCallback, key_cb, key_token);

	XtAddCallback(textout, XtNmouseCallback, mouse_cb, mouse_token);

	XtAddCallback(textout, XtNpasteCallback, paste_cb, paste_token);

	/*
	 * Set resources in the application shell to set the window
	 * title and icon name.
	 */

	{
	    char *icon_prefix = "T:";

	    M0buf *icon_title;
	    M0buf *win_title;

	    icon_title = M0bufalloc();
	    win_title = M0bufalloc();

	    M0bufcat(icon_title, icon_prefix, strlen(icon_prefix));
	    M0window_titles(win_title, icon_title);

	    XtVaSetValues(toplevel,
			XtNiconName, M0bufptr(icon_title),
			XtNtitle, M0bufptr(win_title),
			NULL);

	    M0buffree(icon_title);
	    M0buffree(win_title);
	}

	/*
	 * Allow the window to see keystrokes when it has the
	 * input focus but the pointer is not in the window.
	 */

	XtSetKeyboardFocus(toplevel, textout);

	/*
	 * Do not map the window immediately upon realizing it.
	 * We wait until the image window has been mapped.
	 */

	XtSetMappedWhenManaged(toplevel, False);

	XtRealizeWidget(toplevel);

	/*
	 * Get the initial dimensions of the widget
	 */

	XtVaGetValues(textout,
			    XtNrows, &TWrows,
			    XtNcolumns, &TWcols,
			    NULL);

	/*
	 * Now toplevel has a window, so we can call XSetWMProtocols
	 * and set the command window ID in UC.  However, we only
	 * set the command window ID in UC if the word in UC is zero.
	 * This way, if you have multiple mctext processes, mcimage
	 * only forwards to the first mctext process.  This makes
	 * sense, since the usual case is that the firct mctext is
	 * a child of mcenv, so when it exits, the whole session
	 * exits.  This is not the case with other mctext processes,
	 * which you could start and exit at will.
	 */
	XSetWMProtocols(XtDisplay(toplevel), XtWindow(toplevel),
			&wm_delete_window, 1);

	if (Mcluc(UC_COMMAND_WINDOW_ID) == 0)
	{
		Mcpuc((Fint)XtWindow(textout), UC_COMMAND_WINDOW_ID);
	}

	return (char *)0;
}

static void TWfini(void)
{
	/*
	 * clear our window ID from UC, if our window ID was in UC
	 */

	if (Mcluc(UC_COMMAND_WINDOW_ID) == (Fint) XtWindow(textout))
	{
		Mcpuc(0, UC_COMMAND_WINDOW_ID);
	}

	/*
	 * close our X connection, destroy windows, etc.
	 */

	XtDestroyApplicationContext(app_context);
}

static void TWgetsize(int *rows, int *cols)
{
	if(rows)
		*rows = TWrows;
	if(cols)
		*cols = TWcols;
}

static void TWaddch(TWchar character)
{
	static int old_attrib = 0;

	int    this_attrib;
	char   this_char;

	int    intensity;
	int    fg_ansi;
	int    bg_ansi;

	/*
	 * If we are using the same attribute as before, just
	 * append the character to the buffer.
	 */

	this_attrib =  (int) TW_COLOR_GET(character);
	this_char   = (char) TW_CHAR_GET(character);

	if (this_attrib == old_attrib)
	{
		M0bufcat(TWbuf, &this_char, sizeof this_char);
		return;
	}

	M0tcolor_to_ansi(this_attrib, &intensity, &fg_ansi, &bg_ansi);

	{
		char	buf[64];

		sprintf(buf, "\033[%d;%d;%dm", intensity,
			fg_ansi, bg_ansi);
		M0bufcat(TWbuf, buf, strlen(buf));
	}

	old_attrib = this_attrib;
	M0bufcat(TWbuf, &this_char, sizeof this_char);
}

static void TWclrtoeol(void)
{
	static char seq[] = "\033[K";
	M0bufcat(TWbuf, seq, sizeof seq - 1);
}

static void TWmove(int y, int x)
{
	char buf[64];

	sprintf(buf, "\033[%d;%dH", y+1, x+1);
	M0bufcat(TWbuf, buf, strlen(buf));
}

static int update_cursor_y = 0;
static int update_cursor_x = 0;

/*
 * Set the position that the cursor is moved to at update.
 *
 * Setting this allows us to move the cursor wherever we want to
 * in various routines, since we don't have to worry about saving
 * and restoring the cursor position.
 */

static void
TWcursorpos(int y, int x)
{
	update_cursor_y = y;
	update_cursor_x = x;
}

static void
TWcursortype(int type)
{
	char buf[64];

	switch (type)
	{
	case EBMODE_VI:
		sprintf(buf, "\033]%dX", XMcTermCursorFullSolid);
		break;
	case EBMODE_INSERT:
		sprintf(buf, "\033]%dX", XMcTermCursorHalfSolid);
		break;
	default: /* EBMODE_CHANGE */
		sprintf(buf, "\033]%dX", XMcTermCursorQuarterSolid);
		break;
	}

	M0bufcat(TWbuf, buf, strlen(buf));
}

static void
TWupdate(void)
{
	TWmove(update_cursor_y, update_cursor_x);
	XMcTermWrite(textout, M0bufptr(TWbuf), M0buflen(TWbuf));
	M0bufrealloc(TWbuf, 0);
}

/*
 * mcterm has the newline glitch
 */

static int
TWdelayed_wrap(void)
{
	return 1;
}

/*
 * emits a beep
 */
static void
TWbeep(int freq, int ms)
{
	Display *disp;
	unsigned long value_mask;
	XKeyboardState old_state;
	XKeyboardControl new_values;

	disp = XtDisplay(textout);

	if (freq <= 0 || ms <= 0)
	{
		XBell(disp, 0);
		return;
	}

	/* save current state */
	XGetKeyboardControl(disp, &old_state);

	/* set the desired parameters and ring the bell */
	value_mask = KBBellPitch | KBBellDuration;
	new_values.bell_pitch = freq;
	new_values.bell_duration = ms;
	XChangeKeyboardControl(disp, value_mask, &new_values);
	XBell(disp, 100);

	/* restore the old bell_pitch and bell_duration */
	new_values.bell_pitch = (int) old_state.bell_pitch;
	new_values.bell_duration = (int) old_state.bell_duration;
	XChangeKeyboardControl(disp, value_mask, &new_values);
}

/*
 * raise the text window
 */
static void
TWraise(void)
{
	XRaiseWindow(XtDisplay(toplevel), XtWindow(toplevel));
}

/*
 * main program loop - never returns
 */

/*
 * The out_cb function is called when there is data available on the
 * command output file descriptor.
 * We use it to call the function supplied by the main program.
 */

static XtInputId out_id;
static void (*out_fun)(void);

static void
out_cb(XtPointer client_data, int *fid, XtInputId *id)
{
	Widget term = (Widget) client_data;

	assert(term == textout);
	assert(*id == out_id);
	assert(*fid != -1);

	out_fun();
}

/*
 * The resize_cb function is called whenever the number of rows or
 * columns in the McTerm widget changes.
 * We use it to update the dimensions of the widget and to
 * call the function supplied by the main program.
 */

static char resize_token[] = "blah";
static void (*resize_fun)(void);

static void
resize_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *token = (char *) client_data;
	XMcTermWindowSizeInfo *size;

	size = (XMcTermWindowSizeInfo *) call_data;

	assert(w);
	assert(token == resize_token);

	TWcols = size->columns;
	TWrows = size->rows;

	resize_fun();
}

/*
 * The work_cb function is called at regular intervals.
 * We use it to call the function supplied by the main program.
 */

static char work_token[] = "blah";
static XtIntervalId work_timer;
static void (*work_fun)(void);
enum {WORK_DELAY = 100};

static void
work_handler(XtPointer client_data, XtIntervalId *id)
{
	char *token = (char *) client_data;
	assert(token == work_token);
	assert(*id == work_timer);

	work_fun();

	if(hitbits & HIT_INTERRUPT)
	{
		TWfini();
		exit(0);
	}

	work_timer = XtAppAddTimeOut(app_context, WORK_DELAY,
				     work_handler, work_token);
}

static void
TWmainloop(int pipefd, void (*pipefdfun)(void), void (*winchfun)(void),
	 void (*idlefun)(void), void (*waitfun)(void))
{

	out_fun = pipefdfun;
	out_id = XtAppAddInput(app_context, pipefd,
		      (XtPointer) XtInputReadMask, out_cb, textout);

	resize_fun = winchfun;
	XtAddCallback(textout, XtNresizeCallback,
			resize_cb, resize_token);

	work_fun = idlefun;
	work_timer = XtAppAddTimeOut(app_context, WORK_DELAY,
				     work_handler, work_token);

	/*
	 * After possibly waiting for the image window to become
	 * visible, we make our window visible.
	 */

	waitfun();
	XtMapWidget(toplevel);

	/*
	 * Main loop.
	 */

	/* XtAppMainLoop(app_context); */
	for(;;)
	{
		XEvent event;

		XtAppNextEvent(app_context, &event);
		M0XKeyUpdate(&event);
		XtDispatchEvent(&event);
	}
}

TWop *
TWwinxop(void)
{
	static TWop op;
	static TWop *pop;

	if (pop)
		return pop;

	/*
	 * hardcoding the number of pointers in a TWop
	 * structure catches cases where we change the
	 * set of pointers in a TWop structure
	 */

	assert(sizeof op == 16 * sizeof(void *));

	op.TWinit		= TWinit;
	op.TWfini		= TWfini;
	op.TWupdate		= TWupdate;
	op.TWbeep		= TWbeep;
	op.TWsuspend		= 0;
	op.TWraise		= TWraise;

	op.TWgetsize		= TWgetsize;
	op.TWaddch		= TWaddch;
	op.TWclrtoeol		= TWclrtoeol;
	op.TWmove		= TWmove;
	op.TWcursorpos		= TWcursorpos;
	op.TWcursortype		= TWcursortype;

	op.TWmainloop		= TWmainloop;

	op.TWnext_font		= TWnext_font;
	op.TWprev_font		= TWprev_font;

	op.TWdelayed_wrap	= TWdelayed_wrap;

	pop = &op;

	return pop;
}

#if 0

/*
 * This code is not currently used, but it is a useful example of how
 * to tell if an X window ID is valid...
 */

/*
** Name:
**	M0is_window - is the given X window ID valid?
**
** Interface:
**	#include <X11/Xlib.h>
**
**	static int
**	M0is_window(Display *disp, Window win)
**
** Input:
**	disp    - X display
**	win     - window ID
**
** Input and Output:
**	none
**
** Output:
**	none
**
** Return values:
**	 1	- true
**	 0	- false
**
** Remarks:
**      This routine returns true if the given windowid is defined
**	on the given X display.
**
**	It is useful for avoiding BadWindow errors.
**
** Categories:
**	none
*/

/* This holds the old error handler that we save & restore */

static int    (*my_old_handler)(Display *, XErrorEvent *);

/* This allows my_handler() to check that any XErrorEvent is
 * for the window id we were interested in */

static XID	my_window;

/* This allows my_handler() to communicate the presence of
 * a BadWindow error to M0is_window() */

static int	my_error_flag;

/* This is the error handler used by M0is_window(). */

static int
my_handler(Display *disp, XErrorEvent *event)
{
	/*
	 * If ListProperties gives us an error,
	 * record it and return.
	 */

	if (event->error_code == BadWindow
	&& event->resourceid == my_window
	&& event->request_code == X_ListProperties)
	{
		my_error_flag = 1;
		return 0;
	}

	/*
	 * Otherwise, invoke the previous error handler.
	 */

	return my_old_handler(disp, event);
}

/*
 * This routine true if the given windowid is defined on the given
 * display.
 *
 * It is useful for avoiding BadWindow errors.
 *
 * It sends a ListProperties request to the given window.
 * This is a good way to test if the window exists since the
 * ListProperties request does not change the window in any way.
 * Also, since the ListProperties request has a reply, any errors
 * will come back before XListProperties() returns.
 *
 * It installs a temporary error handler to tell if there is a
 * BadWindow XError from the ListProperties request.
 */

static int
M0is_window(Display *disp, Window win)
{
	int nlist;
	Atom *list;

	my_window = win;
	my_error_flag = 0;
	my_old_handler = XSetErrorHandler(my_handler);

	list = XListProperties(disp, win, &nlist);

	(void) XSetErrorHandler(my_old_handler);

	if (list)
		XFree(list);

	return ! my_error_flag;
}

#endif
